"""
The datasets module has only one function.  It could probably be refactored
into the Triangle class, and may do so in the future.

"""
from pandas import read_pickle
import os


def load_dataset(key):
    """ Function to load datasets included in the chainladder package.
   
    Arguments:
	key: str
	    The name of the dataset, e.g. RAA, ABC, UKMotor, GenIns, etc.
    
    Returns:
	pandas.DataFrame of the loaded dataset.
   """
    path = os.path.dirname(os.path.abspath(__file__))
    return read_pickle(os.path.join(path, 'data', key))

    



# -*- coding: utf-8 -*-

"""The MackChainladder module and class establishes the statistical framework
of the chainladder method.  With MackChainLadder, various standard errors can
be computed to ultimately develop confidence intervals on IBNR estimates.   
"""
import numpy as np
from pandas import DataFrame, concat, Series
from scipy import stats
from chainladder.Chainladder import Chainladder
from warnings import warn
import matplotlib.pyplot as plt
import seaborn as sns
from statsmodels.nonparametric.smoothers_lowess import lowess


class MackChainladder:
    """ MackChainLadder class specifies the Mack chainladder model.

    Thomas Mack published in 1993 [Mac93] a method which estimates the standard
    errors of the chain-ladder forecast without assuming a distribution under 
    three conditions.
    `Proper Citation Needed... <https://github.com/mages/ChainLadder>`_

    Parameters:    
        tri : Triangle
            A triangle object. Refer to :class:`Classes.Triangle`
        weights : int
            A value representing an input into the weights of the WRTO class.
        alpha : int
            A value representing an input into the weights of the WRTO class.
        tail : bool
            Represent whether a tail factor should be applied to the data. 
            Value of False sets tail factor to 1.0

    Attributes:
        tri : Triangle
            A triangle object. Refer to :class:`Classes.Triangle`
        weights : pandas.DataFrame
            A value representing an input into the weights of the WRTO class.
        full_triangle : pandas.DataFrame
            A completed triangle using Mack Chainladder assumptions.
        f : numpy.array
            An array representing the (col-1) loss development factors, 
            f-notation borrowed from Mack
        fse : numpy.array
            An array representing the (col-1) standard errors of loss 
            development factors.

    """

    def __init__(self, tri,  weights=1,  alpha=1,
                 tail=False):
        # Determine whether LDFs can eb extrapolated with exponential tail
        if tail == True:
            tail = self.is_exponential_tail_appropriate(tri,  weights,  alpha) 
            
        self.chainladder = Chainladder(tri, weights=weights,  delta=2 - alpha, tail=tail)
        self.triangle = self.chainladder.triangle
        self.alpha = [2 - item for item in self.chainladder.delta]
        self.weights = self.chainladder.weights
        self.full_triangle = self.chainladder.full_triangle
        self.f = self.chainladder.LDF
        self.sigma = np.array([item.sigma for item in self.chainladder.models])[:-1]
        self.sigma = np.append(self.sigma,self.get_tail_sigma())
        self.fse = np.array([item.standard_error for item in self.chainladder.models])[:-1]
        self.fse = np.append(self.fse,self.get_tail_se())
        self.Fse = self.get_Fse()
        self.total_process_risk = np.array(np.sqrt((self.get_process_risk()**2).sum()))
        self.total_parameter_risk = self.get_total_parameter_risk()
        self.mack_se = np.sqrt(self.get_process_risk()**2 +
                               self.get_parameter_risk()**2)
        self.total_mack_se = np.sqrt(
            self.total_process_risk[-1]**2 + self.total_parameter_risk[-1]**2)

    def __repr__(self):   
        return str(self.summary())
    
    def get_process_risk(self):
        """ Method to return the process risk of the Mack Chainladder model.

        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$Mack.ProcessRisk.
        """

        procrisk = DataFrame([0 for item in range(len(self.full_triangle))],
                             index=self.full_triangle.index, columns=[self.full_triangle.columns[0]])
        bool_df = self.triangle.data.isnull()
        bool_df['Ult'] = True
        ind = 0 if self.chainladder.tail == False else 1
        for i in range(1, len(self.triangle.data.columns) + ind):
            temp = DataFrame(np.sqrt((self.full_triangle.iloc[:, i - 1] * self.Fse.iloc[:, i - 1])**2 + (
                self.f[i - 1] * procrisk.iloc[:, i - 1])**2) * bool_df.iloc[:, i])
            temp.columns = [self.full_triangle.columns[i]]
            procrisk = concat([procrisk, temp], axis=1)
        return procrisk

    def get_parameter_risk(self):
        """ Method to return the parameter risk of the Mack Chainladder model.

        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$Mack.ParameterRisk.
        """

        paramrisk = DataFrame([0 for item in range(len(self.full_triangle))],
                              index=self.full_triangle.index, columns=[self.full_triangle.columns[0]])
        bool_df = self.triangle.data.isnull()
        bool_df['Ult'] = True
        ind = 0 if self.chainladder.tail == False else 1
        for i in range(1, len(self.triangle.data.columns)+ind):
            temp = DataFrame(np.sqrt((self.full_triangle.iloc[:, i - 1] * self.fse[i - 1])**2 + (
                self.f[i - 1] * paramrisk.iloc[:, i - 1])**2) * bool_df.iloc[:, i])
            temp.columns = [self.full_triangle.columns[i]]
            paramrisk = concat([paramrisk, temp], axis=1)
        return paramrisk

    def get_total_parameter_risk(self):
        """ Method to produce the parameter risk of the Mack Chainladder model
        for each development column.

        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$Total.ProcessRisk.

        """
        M = np.empty(0)
        tpr = [0]
        ind = 1 if self.chainladder.tail == False else 0
        for i in range(len(self.full_triangle.columns)-ind):
            M = np.append(M, np.array(
                sum(self.full_triangle.iloc[:, i].iloc[-(i + 1):])))
        for i in range(len(self.full_triangle.columns) - 1 - ind):
            tpr.append(np.sqrt((M[i] * self.fse[i]) **
                               2 + (tpr[-1] * self.f[i])**2))
        return np.array(tpr)

    def get_Fse(self):
        """ Method to produce the Full Triangle standard error of the Mack Chainladder 
        model.


        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$F.se
        """

        # This is sloppy, and I don't know that it works for non-uniform
        # weights.
        full_triangle = self.full_triangle.iloc[:,:-1]
        full_triangleweightconst = self.weights.data.mode().T.mode().iloc[0, 0]
        full_triangleweight = self.full_triangle * 0 + full_triangleweightconst
        Fse = DataFrame()
        for i in range(self.full_triangle.shape[1] - 1):
            Fse = concat([Fse, DataFrame(self.sigma[i] /
                                         np.sqrt(np.array(full_triangleweight.iloc[:, i]).astype(float)
                                                 * np.array(self.full_triangle.iloc[:, i]).astype(float)**self.alpha[i]))], axis=1)
        Fse.set_index(self.full_triangle.index, inplace=True)
        Fse.columns = self.triangle.data.columns
        return Fse

    def get_tail_se(self):
        """ Method to produce the standard error of the Mack Chainladder 
        model tail factor


        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$tail.se
        """
        
        tailse = np.array(self.sigma[-2] / \
            np.sqrt(self.full_triangle.iloc[0, -3]**self.alpha[-1]))
        if self.chainladder.tail == True:
            time_pd = self.get_tail_weighted_time_period()
            fse = np.append(self.fse, tailse)
            x = np.array([i + 1 for i in range(len(fse))])
            fse_reg = stats.linregress(x, np.log(fse))
            tailse = np.append(tailse, np.exp(time_pd * fse_reg[0] + fse_reg[1]))
        else:
            tailse = np.append(tailse,0)
        return tailse

    def get_tail_sigma(self):
        """ Method to produce the sigma of the Mack Chainladder 
        model tail factor


        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$tail.sigma
        """
        y = np.log(self.sigma[:len(self.triangle.data.columns[:-2])])
        x = np.array([i + 1 for i in range(len(y))])
        model = stats.linregress(x, y)
        tailsigma = np.exp((x[-1] + 1) * model[0] + model[1])
        if model[3] > 0.05:  # p-vale of slope parameter
            y = self.sigma
            tailsigma = np.sqrt(
                abs(min((y[-1]**4 / y[-2]**2), min(y[-2]**2, y[-1]**2))))
        if self.chainladder.tail == True: 
            time_pd = self.get_tail_weighted_time_period()
            y = np.log(np.append(self.sigma,tailsigma))
            x = np.array([i + 1 for i in range(len(y))])
            sigma_reg = stats.linregress(x, y)
            tailsigma = np.append(tailsigma, np.exp(time_pd * sigma_reg[0] + sigma_reg[1]))
        else:
            tailsigma = np.append(tailsigma,0)
        return tailsigma

    def get_tail_weighted_time_period(self):
        """ Method to approximate the weighted-average development age assuming
        exponential tail fit.
        
        Returns: float32
        """
        n = len(self.triangle.data.columns)-1
        y = self.f[:n]
        x = np.array([i + 1 for i in range(len(y))]) 
        ldf_reg = stats.linregress(x, np.log(y - 1))
        time_pd = (np.log(self.f[n] - 1) - ldf_reg[1]) / ldf_reg[0]
        return time_pd
        
    def summary(self):
        """ Method to produce a summary table of of the Mack Chainladder 
        model.

        Returns:
            This calculation is consistent with the R calculation 
            MackChainLadder$summary
        """
        summary = DataFrame()
        summary['Latest'] = Series([row.dropna(
        ).iloc[-1] for index, row in self.triangle.data.iterrows()], index=self.triangle.data.index)
        summary['Dev to Date'] = Series([row.dropna().iloc[-1] for index, row in self.triangle.data.iterrows(
        )], index=self.triangle.data.index) / self.full_triangle.iloc[:, -1]
        summary['Ultimate'] = self.full_triangle.iloc[:, -1]
        summary['IBNR'] = summary['Ultimate'] - summary['Latest']
        summary['Mack S.E.'] = self.mack_se.iloc[:, -1]
        summary['CV(IBNR)'] = summary['Mack S.E.'] / summary['IBNR']
        return summary
    
    def age_to_age(self, colname_sep='-'):
        """ Simple method that calls on the Chainladder class method of the 
        same name.  See age_to_age() in Chainladder module.

        Parameters:    
            colname_sep : str
                text to join the names of two adjacent columns representing the
                age-to-age factor column name.

        Returns:
            Pandas.DataFrame of the age-to-age triangle.
        """
        return self.chainladder.age_to_age(colname_sep='-')
    
    def is_exponential_tail_appropriate(self, tri,  weights,  alpha):
        """ Method to determine whether an exponential tail fit is appropriate
        based purely on the p-value of the slope parameter.  This method
        currently uses scipy.stats, but this dependency can be removed as 
        statsmodels.api has a more robust regression framework and is used
        elsewhere in the package.
        
        Arguments:
            tri: Triangle 
                An object of the Triangle class that will be used to test the 
                exponential tail fit.
            weights: Triangle 
                An object representing the weights of the exponential regression
                model.
            alpha: pandas.Series
                An object representing the alpha parameter of the mackChainLadder
                model.
        
        Returns:
            Boolean value is returned as True if exponential tail is appropriate
            otherwise a value of False is returned.
                
        
        """
        y = Series(Chainladder(tri, weights=weights,  delta=2 - alpha).LDF)
        y.index = [i+1 for i in range(len(y))]
        y = y[y>1]
        model = stats.linregress(y.index, np.log(y-1))
        if model[3]>0.05:
            warn('Unable to generate tail from LDFs, setting tail factor to 1.0')        
            return False
        else:
            return True
    
    def plot(self, plots=['summary', 'full_triangle', 'resid1', 'resid2','resid3','resid4']): 
        """ Method, callable by end-user that renders the matplotlib plots 
        based on the configurations in __get_plot_dict().
        
        Arguments:
            plots: list[str]
                A list of strings representing the charts the end user would like
                to see.  Of ommitted, all plots are displayed.

        Returns:
            Renders the matplotlib plots.
        """        
        
        sns.set_style("whitegrid")
        _ = plt.figure()
        grid_x = 1 if len(plots) == 1 else round(len(plots) / 2,0)
        grid_y = 1 if len(plots) == 1 else 2
        fig, ax = plt.subplots(figsize=(grid_y*15, grid_x*10))
        for i in range(len(plots)):
            _ = plt.subplot(grid_x,grid_y,i+1)
            self.dict_plot(plots[i]) 

    def dict_plot(self, chart):
        """ Method that renders the matplotlib plots based on the configurations
        in __get_plot_dict().  This method should probably be private and may be
        so in a future release.
        
        """
        my_dict = self.__get_plot_dict()[chart]
        for i in range(len(my_dict['chart_type_dict']['type'])):
            if my_dict['chart_type_dict']['type'][i] == 'plot':
                _ = plt.plot(my_dict['chart_type_dict']['x'][i], 
                             my_dict['chart_type_dict']['y'][i],
                             linestyle=my_dict['chart_type_dict']['linestyle'][i], 
                             marker=my_dict['chart_type_dict']['marker'][i], 
                             color=my_dict['chart_type_dict']['color'][i])
            if my_dict['chart_type_dict']['type'][i] == 'bar':
                _ =plt.bar(height = my_dict['chart_type_dict']['height'][i], 
                           left = my_dict['chart_type_dict']['left'][i],
                           yerr = my_dict['chart_type_dict']['yerr'][i], 
                           bottom = my_dict['chart_type_dict']['bottom'][i])
            if my_dict['chart_type_dict']['type'][i] == 'line':
                _ = plt.gca().set_prop_cycle(None)
                for j in range(my_dict['chart_type_dict']['rows'][i]):
                    _ = plt.plot(my_dict['chart_type_dict']['x'][i], 
                                 my_dict['chart_type_dict']['y'][i].iloc[j], 
                                 linestyle=my_dict['chart_type_dict']['linestyle'][i], 
                                 linewidth=my_dict['chart_type_dict']['linewidth'][i],
                                 alpha=my_dict['chart_type_dict']['alpha'][i])
            _ = plt.title(my_dict['Title'], fontsize=30)

    def __get_plot_dict(self):
        """ Private method that is designed to configure the matplotlib graphs.
        
        Returns:
            Returns a dictionary containing the configuration of the selected plot.
        """
        resid_df = self.chainladder.get_residuals()  
        # Jittering values so LOWESS can calculate
        xlabs = self.full_triangle.columns
        xvals = [i for i in range(len(self.full_triangle.columns))]
        
        summary = self.summary()
        means = list(summary['Ultimate'])
        ci = list(zip(summary['Ultimate']+summary['Mack S.E.'], summary['Ultimate']-summary['Mack S.E.']))
        y_r = [list(summary['Ultimate'])[i] - ci[i][1] for i in range(len(ci))]
        plot_dict = {'resid1':{'Title':'Studentized Residuals by fitted Value',
                             'XLabel':'Fitted Value',
                             'YLabel':'Studentized Residual',
                             'chart_type_dict':{'type':['plot','plot'],
                                               'x':[resid_df['fitted_value'],lowess(resid_df['standard_residuals'],resid_df['fitted_value'],frac=1 if len(np.unique(resid_df['fitted_value'].values))<=6 else 0.666).T[0]],
                                               'y':[resid_df['standard_residuals'],lowess(resid_df['standard_residuals'],resid_df['fitted_value'],frac=1 if len(np.unique(resid_df['fitted_value'].values))<=6 else 0.666).T[1]],
                                               'marker':['o',''],
                                               'linestyle':['','-'],
                                               'color':['blue','red']
                                               }},
                    'resid4':{'Title':'Studentized Residuals by Development Period',
                                     'XLabel':'Development Period',
                                     'YLabel':'Studentized Residual',
                                     'chart_type_dict':{'type':['plot','plot'],
                                                       'x':[resid_df['dev'],lowess(resid_df['standard_residuals'],resid_df['dev'],frac=1 if len(np.unique(resid_df['dev'].values))<=6 else 0.666).T[0]],
                                                       'y':[resid_df['standard_residuals'],lowess(resid_df['standard_residuals'],resid_df['dev'],frac=1 if len(np.unique(resid_df['dev'].values))<=6 else 0.666).T[1]],
                                                       'marker':['o',''],
                                                       'linestyle':['','-'],
                                                       'color':['blue','red']
                                                       }},
                    'resid2':{'Title':'Studentized Residuals by Origin Period',
                                     'XLabel':'Origin Period',
                                     'YLabel':'Studentized Residual',
                                     'chart_type_dict':{'type':['plot','plot'],
                                                       'x':[resid_df.index,lowess(resid_df['standard_residuals'],resid_df.index, frac=1 if len(np.unique(resid_df.index.values))<=6 else 0.666).T[0]],
                                                       'y':[resid_df['standard_residuals'],lowess(resid_df['standard_residuals'],resid_df.index, frac=1 if len(np.unique(resid_df.index.values))<=6 else 0.666).T[1]],
                                                       'marker':['o',''],
                                                       'linestyle':['','-'],
                                                       'color':['blue','red']
                                                       }},
                    'resid3':{'Title':'Studentized Residuals by Origin Period',
                                     'XLabel':'Origin Period',
                                     'YLabel':'Studentized Residual',
                                     'chart_type_dict':{'type':['plot','plot'],
                                                       'x':[resid_df['cal_period'],lowess(resid_df['standard_residuals'],resid_df['cal_period'],frac=1 if len(np.unique(resid_df['cal_period'].values))<=6 else 0.666).T[0]],
                                                       'y':[resid_df['standard_residuals'],lowess(resid_df['standard_residuals'],resid_df['cal_period'],frac=1 if len(np.unique(resid_df['cal_period'].values))<=6 else 0.666).T[1]],
                                                       'marker':['o',''],
                                                       'linestyle':['','-'],
                                                       'color':['blue','red']
                                                       }},      
                    'summary':{'Title':'Ultimates by origin period +/- 1 std. err.',
                                     'XLabel':'Origin Period',
                                     'YLabel':'Ultimates',
                                     'chart_type_dict':{'type':['bar','bar'],
                                                       'height':[summary['Latest'],summary['IBNR']],
                                                       'left':[summary.index,summary.index],
                                                       'bottom':[0,summary['Latest']],
                                                       'yerr':[0,y_r]
                                                       }},
                    'full_triangle':{'Title':'Fully Developed Triangle',
                                     'XLabel':'Origin Period',
                                     'YLabel':'Values',
                                     'chart_type_dict':{'type':['line','line'],
                                                       'rows':[len(self.full_triangle), len(self.triangle.data)],
                                                       'x':[xvals, xvals[:-1]],
                                                       'y':[self.full_triangle, self.triangle.data],
                                                       'linestyle':['--', '-'],
                                                       'linewidth':[5, 5],
                                                       'alpha':[0.5, 1.0]
                                                       
                                                       }} ,
                    'triangle':{'Title':'Latest Triangle Data',
                                     'XLabel':'Origin Period',
                                     'YLabel':'Values',
                                     'chart_type_dict':{'type':['line'],
                                                       'rows':[len(self.triangle.data)],
                                                       'x':[xvals[:-1]],
                                                       'y':[self.triangle.data],
                                                       'linestyle':['-'],
                                                       'linewidth':[5],
                                                       'alpha':[1]
                                                       }} 
                    }
        return plot_dict
    



# -*- coding: utf-8 -*-
"""
The Triangle Module includes the Triangle class.  This base structure of the
class is a pandas.DataFrame.  
"""

from pandas import DataFrame, concat, pivot_table


class Triangle():
    """Triangle class is the basic data representation of an actuarial triangle.

    Historical insurance data is often presented in form of a triangle structure, showing
    the development of claims over time for each exposure (origin) period. An origin
    period could be the year the policy was written or earned, or the loss occurrence
    period. Of course the origin period doesn’t have to be yearly, e.g. quarterly or
    6
    monthly origin periods are also often used. The development period of an origin
    period is also called age or lag. Data on the diagonals present payments in the
    same calendar period. Note, data of individual policies is usually aggregated to
    homogeneous lines of business, division levels or perils.
    Most reserving methods of the ChainLadderpackage expect triangles as input data
    sets with development periods along the columns and the origin period in rows. The
    package comes with several example triangles. 
    `Proper Citation Needed... <https://github.com/mages/ChainLadder>`_

    Parameters:   
        data : pandas.DataFrame
            A DataFrame representing the triangle
        origin : numpy.array or pandas.Series
            An array representing the origin period of the triangle.
        dev : numpy.array or pandas.Series
            An array representing the development period of the triangle. In triangle form,
            the development periods must be the columns of the dataset
        values : str
            A string representing the column name of the triangle measure if the data is in
            tabular form.  Otherwise it is ignored.
        dataform : str
            A string value that takes on one of two values ['triangle' and 'tabular']

    Attributes:
        data : pandas.DataFrame
            A DataFrame representing the triangle
        origin : numpy.array or pandas.Series
            Refer to parameter value.
        dev : numpy.array or pandas.Series
            Refer to parameter value.
        values : str
            Refer to parameter value.
        dataform : str
            Refer to parameter value.

    """

    def __init__(self, data=None, origin=None, dev=None, values=None, dataform='triangle'):
        # Currently only support pandas dataframes as data
        if type(data) is not DataFrame:
            raise TypeError(str(type(data)) + ' is not a valid datatype for the Triangle class.  Currently only DataFrames are supported.')
        self.data = data
        self.origin = origin
        if origin == None:
            origin_in_col_bool, origin_in_index_bool = self.__set_origin()
        self.dev = dev
        if dev == None:
            dev_in_col_bool = self.__set_dev()
        self.dataform = dataform
        if dev_in_col_bool == True and origin_in_col_bool == True:
            self.dataform = 'tabular'
        self.values = values
        if values == None and self.dataform == 'tabular':
            self.__set_values()
        #self.latest_values = Series(
        #    [row.dropna().iloc[-1] for index, row in self.data.iterrows()], index=self.data.index)
        
    def __repr__(self):   
        return str(self.data)
    
    def data_as_table(self, inplace=False):
        """Method to convert triangle form to tabular form.

        Arguments:
            inplace: bool
                Set to True will update the instance data attribute inplace 

        Returns:
            Updated instance `data` parameter if inplace is set to True otherwise it returns a pandas.DataFrame
        """
        lx = DataFrame()
        if self.dataform == 'triangle':
            for val in range(len(self.data.T.index)):
                df = DataFrame(self.data.iloc[:, val].rename('values'))
                df['dev'] = int(self.data.iloc[:, val].name)
                lx = lx.append(df)
            lx.dropna(inplace=True)
            if inplace == True:
                self.data = lx[['dev', 'values']]
                self.dataform = 'tabular'
                self.dev = 'dev'
                self.values = 'values'
            return lx[['dev', 'values']]
        else:
            return self.data

    def data_as_triangle(self, inplace=False):
        """Method to convert tabular form to triangle form.
    
        Arguments:
            inplace: bool
                Set to True will update the instance data attribute inplace 
    
        Returns:
            Updated instance `data` parameter if inplace is set to True otherwise it returns a pandas.DataFrame
        """
        if self.dataform == 'tabular':
            tri = pivot_table(self.data, values=self.values, index=[
                              self.origin], columns=[self.dev]).sort_index()
            tri.columns = [str(item) for item in tri.columns]
            if inplace == True:
                self.data = tri
                self.dataform = 'triangle'
            return tri
        else:
            return self.data

    def incr_to_cum(self, inplace=False):
        """Method to convert an incremental triangle into a cumulative triangle.  Note,
        the triangle must be in triangle form.

        Arguments:
            inplace: bool
                Set to True will update the instance data attribute inplace 

        Returns:
            Updated instance `data` parameter if inplace is set to True otherwise it returns a pandas.DataFrame

        """
        incr = DataFrame(self.data.iloc[:, 0])
        for val in range(1, len(self.data.T.index)):
            incr = concat(
                [incr, self.data.iloc[:, val] + incr.iloc[:, -1]], axis=1)
        incr = incr.rename_axis('dev', axis='columns')
        incr.columns = self.data.T.index
        if inplace == True:
            self.data = incr
        return incr

    def cum_to_incr(self, inplace=False):
        """Method to convert an cumulative triangle into a incremental triangle.  Note,
        the triangle must be in triangle form.

        Arguments:
            inplace: bool
                Set to True will update the instance data attribute inplace 

        Returns:
            Updated instance `data` parameter if inplace is set to True otherwise it returns a pandas.DataFrame

        """
        incr = self.data.iloc[:, 0]
        for val in range(1, len(self.data.T.index)):
            incr = concat([incr, self.data.iloc[:, val] -
                           self.data.iloc[:, val - 1]], axis=1)
        incr = incr.rename_axis('dev', axis='columns')
        incr.columns = self.data.T.index
        if inplace == True:
            self.data = incr
        return incr

    def __set_origin(self):
        """Experimental hidden method. Purpose is to profile the data and autodetect the origin period
        improving the user experience by not requiring the user to supply an explicit origin.

        """
        origin_names = ['accyr', 'accyear', 'accident year', 'origin', 'accmo', 'accpd',
                        'accident month']
        origin_in_col_bool = False
        origin_in_index_bool = False
        origin_in_index_T_bool = False
        origin_match = [i for i in origin_names if i in self.data.columns]
        if len(origin_match) == 1:
            self.origin = origin_match[0]
            origin_in_col_bool = True
        if len(origin_match) == 0:
            # Checks for common origin names in dataframe index
            origin_match = [
                i for i in origin_names if i in self.data.index.name]
            if len(origin_match) == 1:
                self.origin = origin_match[0]
                origin_in_index_bool = True
        return origin_in_col_bool, origin_in_index_bool

    def __set_dev(self):
        """Experimental hidden method. Purpose is to profile the data and autodetect the dev period
        improving the user experience by not requiring the user to supply an explicit dev.

        """
        ##### Identify dev Profile ####
        dev_names = ['devpd', 'dev', 'development month', 'devyr', 'devyear']
        dev_in_col_bool = False
        dev_in_index_bool = False
        dev_in_index_T_bool = False
        dev_match = [i for i in dev_names if i in self.data.columns]
        if len(dev_match) == 1:
            self.dev = dev_match[0]
            dev_in_col_bool = True
        return dev_in_col_bool

    def __set_values(self):
        """Experimental hidden method. Purpose is to profile the data and autodetect the values parameter
        improving the user experience by not requiring the user to supply an explicit values parameter.
        This is onyl necessary when dataform is 'tabular'.

        """
        ##### Identify values Profile ####
        value_names = ['incurred claims']
        values_match = [i for i in value_names if i in self.data.columns]
        if len(values_match) == 1:
            self.values = values_match[0]
        else:
            self.values = 'values'
        return



# -*- coding: utf-8 -*-
"""
Created on Sat Jun 17 07:08:08 2017

@author: jboga
"""


from chainladder.UtilityFunctions import load_dataset
from chainladder.Chainladder import Chainladder, WRTO
from chainladder.Triangle import Triangle
from chainladder.MackChainladder import MackChainladder




# -*- coding: utf-8 -*-
"""
The Chainladder Module allows for the computation of basic chainladder methods 
on an object of the Triangle class.  The module contains two classes, 
ChainLadder and WRTO.  Chainladder is the main class while WRTO (Weighted 
regression through the origin) is a unique case of a linear model with 0 
intercept and a single slope parameter which is the form of a chainladder LDF
approach.
"""

import numpy as np
from pandas import DataFrame, concat, Series
import statsmodels.api as sm
from statsmodels.stats.outliers_influence import OLSInfluence
from warnings import warn
from chainladder.Triangle import Triangle

class Chainladder():
    """ ChainLadder class specifies the chainladder model.

    The classical chain-ladder is a deterministic algorithm to forecast claims 
    based on historical data. It assumes that the proportional developments of 
    claims from one development period to the next are the same for all origin 
    years.

    Mack uses alpha between 0 and 2 to distinguish
    alpha = 0 straight averages
    alpha = 1 historical chain ladder age-to-age factors
    alpha = 2 ordinary regression with intercept 0

    However, in Zehnwirth & Barnett they use the notation of delta, whereby 
    delta = 2 - alpha the delta is than used in a linear modelling context.

    `Need to properly cite... <https://github.com/mages/ChainLadder>`_

    Parameters:    
        tri : Triangle
            A triangle object. Refer to :class:`Classes.Triangle`
        weights : int
            A value representing an input into the weights of the WRTO class.
        delta : int
            A value representing an input into the weights of the WRTO class.
        tail : bool
            A value representing whether the user would like an exponential tail
            factor to be applied to the data.

    Attributes:
        tri : Triangle
            A triangle object on which the Chainladder model will be built.
        weights : pandas.DataFrame
            A value representing an input into the weights of the WRTO class.
        delta : list
            A set of values representing an input into the weights of the WRTO 
            class.
        models : list
            A list of WTRO objects for of length (col-1)
        tail : bool
            A value representing whether the user would like an exponential tail
            factor to be applied to the data.
        LDF : pandas.Series
            A series containing the LDFs of the chainladder model.  If tail=True
            and tail fitting succeeds, an additional 'Age-Ult' factor is appended
            the the end representing the tail CDF.
        CDF : pandas.Series
            A series representing the cumulative development factors of the
            chainladder model.
        full_triangle : Pandas.DataFrame
            A table representing the raw triangle data as well as future
            lags populated with the expectation from the chainladder fit.

    """

    def __init__(self, tri, weights=1, delta=1, tail=False):
        if type(tri) is Triangle:
            self.triangle = tri 
        else:
            self.triangle = Triangle(tri)
        self.delta = [delta for item in range(len(self.triangle.data.columns) )]
        self.weights = Triangle(self.triangle.data * 0 + weights)
        self.models = self.fit()
        self.tail = tail
        self.LDF = self.get_LDF().append(self.get_tail_factor())
        self.CDF = self.LDF[::-1].cumprod()[::-1]
        self.full_triangle = self.predict()
        
    def __repr__(self):   
        return str(self.age_to_age())
    
    def predict(self):
        """ Method to 'square' the triangle based on the WRTO 'models' list.
        
        Returns:
            pandas.DataFrame representing the raw triangle data as well as future
            lags populated with the expectation from the chainladder fit.
        """
        ldf = [item.coefficient for item in self.models]
        square = self.triangle.data.copy()
        for row in range(1, len(square)):
            for col in range(row, 0, -1):
                square.iloc[row, -col] = square.iloc[row, -col - 1] * ldf[-col]
        square['Ult']=square.iloc[:,-1]*self.CDF[-1]
        return square

    def fit(self):
        """ Method to call the weighted regression trhough the origin fitting.
        
        Returns:
            list of statsmodel (Chainladder) models with a subset of properties 
            that can be accessed later.  See WRTO class implementation.
        """
        models = []
        for i in range(0, len(self.triangle.data.columns) - 1):
            data = self.triangle.data.iloc[:, i:i + 2].dropna()
            w = self.weights.data.iloc[:, i].dropna().iloc[:-1]
            X = data.iloc[:, 0]
            y = data.iloc[:, 1]
            sample_weight = np.array(w).astype(
                float) / np.array(X).astype(float)**self.delta[i]
            lm = WRTO(X, y, sample_weight)
            models.append(lm)
        return models

    def age_to_age(self, colname_sep='-'):
        """ Method to display an age-to-age triangle with a display of simple 
        average chainladder development factors and volume weighted average 
        development factors.

        Parameters:    
            colname_sep : str
                text to join the names of two adjacent columns representing the
                age-to-age factor column name.

        Returns:
            Pandas.DataFrame of the age-to-age triangle.
        """
        incr = DataFrame(self.triangle.data.iloc[:, 1] /
                         self.triangle.data.iloc[:, 0])
        for i in range(1, len(self.triangle.data.columns) - 1):
            incr = concat([incr, self.triangle.data.iloc[:, i + 1] /
                           self.triangle.data.iloc[:, i]], axis=1)
        incr.columns = [item + colname_sep +
                        self.triangle.data.columns.values[num + 1] for num,
                        item in enumerate(self.triangle.data.columns.values[:-1])]
        incr = incr.iloc[:-1]
        incr[str(self.triangle.data.columns.values[-1]) + colname_sep + 'Ult'] = np.nan
        ldf = Chainladder(self.triangle.data, delta=2, tail=self.tail).LDF
        incr.loc['simple'] = ldf
        ldf = Chainladder(self.triangle.data, delta=1, tail=self.tail).LDF
        incr.loc['vol-wtd'] = ldf
        incr.loc['Selected'] = ldf
        ldf = self.LDF
        return incr
    
    def get_LDF(self, colname_sep='-'):
        """ Method to obtain the loss development factors (LDFs) from the
        chainladder model.
        
        Parameters:    
            colname_sep : str
                text to join the names of two adjacent columns representing the
                age-to-age factor column name.

        Returns:
            Pandas.Series of the LDFs.
        
        """
        LDF = Series([ldf.coefficient for ldf in self.models], index=[item + colname_sep +
                        self.triangle.data.columns.values[num + 1] for num,
                        item in enumerate(self.triangle.data.columns.values[:-1])])
        if len(LDF) >= len(self.triangle.data.columns)-1:
            return Series(LDF[:len(self.triangle.data.columns)-1])
        else:
            return LDF
            
         
        

    def get_tail_factor(self, colname_sep='-'):
        """Estimate tail factor, idea from Thomas Mack:
        Returns a tail factor based off of an exponential fit to the LDFs.  This will
        fail if the product of 2nd and 3rd to last LDF < 1.0001.  This also fails if
        the estimated tail is larger than 2.0.  In other areas, exponential fit is
        rejected if the slope parameter p-value >0.5.  This is currently representative
        of the R implementation of this package, but may be enhanced in the future to be
        p-value based.

        Parameters:    
            colname_sep : str
                text to join the names of two adjacent columns representing the
                age-to-age factor column name.

        Returns:
            Pandas.Series of the tail factor.        
        """
        LDF = self.get_LDF()[:len(self.triangle.data.columns)-1]
        n = len(LDF)
        if (LDF[-1] * LDF[-2] > 1.0001) and self.tail == True:
            LDF_positive = np.array([item for item in LDF if item > 1])
            n = np.max(np.where(LDF >= 1)[0])            
            tail_model = np.polyfit(range(1,len(LDF_positive)+1),np.log(LDF_positive-1),1)
            tail_factor = np.product(np.exp(tail_model[1] + np.array([i+2 for i in range(n,n+100)]) * tail_model[0]).astype(float) + 1)
            if tail_factor > 2:
                warn("The estimate tail factor was bigger than 2 and has been reset to 1.")
                tail_factor = 1
        else:
            tail_factor = 1
        return Series(tail_factor, index = [self.triangle.data.columns[-1] + colname_sep + 'Ult'])

    def get_residuals(self):
        """Generates a table of chainladder residuals along with other statistics.
        These get used in the MackChainLadder.plot() method for residual plots.
         
        Returns:
            Pandas.DataFrame of the residual table        
        """

        Resid = DataFrame()
        for i in range(len(self.models)):
            resid = DataFrame()
            resid['x'] = self.models[i].x
            resid['dev'] = int(self.models[i].x.name)
            resid['cal_period'] = resid.index + resid['dev'] - 1
            resid['residuals'] = self.models[i].residual
            resid['standard_residuals'] = self.models[i].std_resid
            resid['fitted_value'] = self.models[i].fittedvalues
            Resid = Resid.append(resid)
        return Resid.drop(['x'], axis=1).dropna()
    
class WRTO():
    """Weighted least squares regression through the origin

    Collecting the relevant variables from statsmodel OLS/WLS. Note in
    release 0.1.0 of chainladder, there is a deprecation warning with statsmodel
    that will persist until statsmodel is upgraded.

    Parameters:    
        X : numpy.array or pandas.Series
            An array representing the independent observations of the 
            regression.
        y : numpy.array or pandas.Series
            An array representing the dependent observations of the regression.
        w : numpy.array or pandas.Series
            An array representing the weights of the observations of the 
            regression.

    Attributes:
        X : numpy.array or pandas.Series
            An array representing the independent observations of the 
            regression.
        y : numpy.array or pandas.Series
            An array representing the dependent observations of the regression.
        w : numpy.array or pandas.Series
            An array representing the weights of the observations of the 
            regression.
        coefficient : numpy.float64
            Slope parameter of the regression.
        WSSResidual : numpy.float64
            Weighted residual sum of squares of the regression.
        mean_square_error : numpy.float64
            Mean square error of the regression.
        residual : numpy.array
            The difference between actual y value and the fitted/expected y value
        standard_error : numpy.float64
            Standard error of the regression slope paramter.
        sigma : numpy.float64
            Square root of the mean square error of the regression. 
        std_resid : numpy.array
            Represents internally studentized residuals which generally vary between
            [-2,2].  Used in residual scatterplots and help determine the appropriateness
            of the model on the data.

    """

    def __init__(self, x, y, w):
        self.x = x
        self.y = y
        self.w = w
        WLS = sm.WLS(y,x, w)
        OLS = sm.OLS(WLS.wendog,WLS.wexog).fit()
        #if OLS.params[0] = 
        self.coefficient = OLS.params[0]
        self.WSSResidual = OLS.ssr
        self.fittedvalues = OLS.predict(x)
        self.residual = OLS.resid
        if len(x) == 1:
            self.mean_square_error = np.nan 
            self.standard_error = np.nan
            self.sigma = np.nan
            self.std_resid = np.nan
        else:
            self.mean_square_error = OLS.mse_resid 
            self.standard_error = OLS.params[0]/OLS.tvalues[0]
            self.sigma = np.sqrt(self.mean_square_error)
            self.std_resid = OLSInfluence(OLS).resid_studentized_internal
        
        
        
        


descr = "Chainladder Package - P&C actuarial package modeled after the R package of the same name"

from distutils.core import setup

setup(
    name='chainladder',
    version='0.1.2',
    maintainer='John Bogaardt',
    maintainer_email='jbogaardt@gmail.com',
    packages=['chainladder'],
    scripts=[],
    url='https://github.com/jbogaardt/chainladder-python',
    download_url='https://github.com/jbogaardt/chainladder-python/archive/v0.1.1.tar.gz',
    license='LICENSE',
    include_package_data=True,
    package_data = {'data':['/data/ABC',
                    '/data/M3IR5',
                    '/data/MCLincurred',
                    '/data/MCLpaid',
                    '/data/Mortgage',
                    '/data/MW2008',
                    '/data/MW2014',
                    '/data/qincurred',
                    '/data/qpaid',
                    '/data/RAA',
                    '/data/UKMotor',
                    '/data/USAAincurred',
                    '/data/USAApaid',
                    '/data/auto$CommercialAutoPaid',
                    '/data/auto$PersonalAutoIncurred',
                    '/data/auto$PersonalAutoPaid',
                    '/data/GenIns',
                    '/data/GenInsLong',
                    '/data/liab$AutoLiab',
                    '/data/liab$GeneralLiab']},
    description= descr,
    #long_description=open('README.rst').read(),
    install_requires=[
        "pandas>=0.20.2",
        "numpy>=1.12.1",
        "matplotlib>=2.0.2",
        "seaborn>=0.7.1",
        "scipy>=0.19.0",
        "statsmodels>=0.8.0",
    ],
)